# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_losses.ipynb.

# %% auto 0
__all__ = ['CombinedLoss']

# %% ../nbs/04_losses.ipynb 3
from fastai.losses import FocalLossFlat, DiceLoss
from fastcore.basics import store_attr

# %% ../nbs/04_losses.ipynb 4
class CombinedLoss:
    "Dice and Focal combined"
    def __init__(self, axis=1, smooth=1.0, alpha=1.0, gamma=2.0):
        store_attr()
        self.focal_loss = FocalLossFlat(gamma = gamma, axis=axis)
        self.dice_loss =  DiceLoss(axis, smooth, reduction='mean')
        
    def __call__(self, pred, targ, reduction=None):
        return (1. - self.alpha) * self.focal_loss(pred, targ) + self.alpha * self.dice_loss(pred, targ)
        #return self.dice_loss(pred, targ)
    
    def decodes(self, x):    return x.argmax(dim=self.axis)
    def activation(self, x): return F.softmax(x, dim=self.axis)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_model.ipynb.

# %% auto 0
__all__ = ['MODEL_DEFS', 'get_seg_model', 'get_preprocessiong_fn', 'get_model_name', 'calc_confusion_matrix',
           'display_confusion_matrix', 'save_model_checkpoint', 'create_checkpoint_data', 'restore_model']

# %% ../nbs/03_model.ipynb 3
import segmentation_models_pytorch as smp
import torch
from torchmetrics.classification import MulticlassConfusionMatrix
import matplotlib.pyplot as plt
import torch.nn.functional as F

# %% ../nbs/03_model.ipynb 5
def get_seg_model(model_def, classes, weights_fn=''):
    """
    Create a segmentation model, and load provided weighst
    """
    ENCODER = model_def['encoder']
    ENCODER_WEIGHTS = model_def['weights']
    ACTIVATION = None
    CLASSES = classes
    model_func = eval(model_def['arch'])
    model = model_func(
        encoder_name = ENCODER,
        encoder_weights=ENCODER_WEIGHTS,
        classes = len(CLASSES),
        activation = ACTIVATION
        )
    if weights_fn != '':
        checkpoint = torch.load(weights_fn)
        model.load_state_dict(checkpoint['model'])
    
    return model

# %% ../nbs/03_model.ipynb 7
def get_preprocessiong_fn(model_def):
    """
    Return the preprocessing function for a specific segmentation model
    """
    ENCODER = model_def['encoder']
    ENCODER_WEIGHTS = model_def['weights']
    return smp.encoders.get_preprocessing_fn(ENCODER, ENCODER_WEIGHTS)

def get_model_name(model_def, run_name=''):
    """
    Return the model name incl. a run name
    """
    arch_name = model_def['arch'].split('.')[-1]
    model_name = arch_name + '_' + model_def['encoder'] + run_name 
    return model_name

# %% ../nbs/03_model.ipynb 11
def calc_confusion_matrix(model, dl, num_classes):
    model.eval()
    metric = MulticlassConfusionMatrix(num_classes=num_classes,normalize='true')
    val = torch.zeros([num_classes,num_classes])
    size = 0
    for b in iter(dl):
        logits = model.predict(b[0])
        preds = torch.argmax(F.softmax(logits, dim=1), dim=1)
        val += metric(preds.to('cpu'), b[1].to('cpu')) * b[0].shape[0]
        size += b[0].shape[0]
    return val/size

def display_confusion_matrix(val, select_classes):
    fig, ax = plt.subplots(1,1, figsize=(7,7))
    MulticlassConfusionMatrix(num_classes=val.shape[0]).plot((val.round(decimals=2)*100.).trunc(),labels=select_classes, ax = ax)

# %% ../nbs/03_model.ipynb 13
def save_model_checkpoint(model, opt, path, **kwargs):
    d = {'model_state_dict' : model.state_dict(), 'optimizer_state_dict' : opt.state_dict()}
    d |= kwargs
    torch.save(d, path)

def create_checkpoint_data(model_def, select_classes, epoch, loss, metric):
    checkpoint_data = { 'arch' : model_def['arch'], 
                       'encoder' : model_def['encoder'],
                       'weights' : model_def['weights'],
                       'select_classes' : select_classes,
                       'epoch' : epoch,
                       'loss' : loss,
                       'metric' : metric
                        }
    return checkpoint_data

def restore_model(path):
    d = torch.load(path, map_location=torch.device('cpu') )
    model_def= {'arch': d['arch'], 'encoder' : d['encoder'], 'weights' : d['weights']}
    model = get_seg_model(model_def, d['select_classes'])
    model.load_state_dict(d['model_state_dict'])
    return model.eval(), d['select_classes'], model_def

# %% ../nbs/03_model.ipynb 15
MODEL_DEFS = []
MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'efficientnet-b5', 'weights': 'imagenet'}) #5 
MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'efficientnet-b7', 'weights': 'imagenet'}) #8
MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'efficientnet-b7', 'weights': 'imagenet'}) #9
MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'efficientnet-b5', 'weights': 'imagenet'}) #9
MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'efficientnet-b5', 'weights': 'imagenet'}) #10
MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'efficientnet-b7', 'weights': 'imagenet'}) #10
MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'xception', 'weights': 'imagenet'}) #11
#MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'resnext50_32x4d', 'weights': 'imagenet'}) #0
#MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'resnext101_32x8d', 'weights': 'imagenet'})#1
#MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'resnet101', 'weights': 'imagenet'})       #2
#MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'resnext50_32x4d', 'weights': 'ssl'}) #3
#MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'resnext101_32x8d', 'weights': 'imagenet'}) #4 
#MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'efficientnet-b6', 'weights': 'imagenet'}) #6 
#MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'resnext101_32x8d', 'weights': 'imagenet'}) #7 
#MODEL_DEFS.append({'arch': 'smp.Unet', 'encoder': 'inceptionresnetv2', 'weights': 'imagenet'}) #12
#MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'inceptionv4', 'weights': 'imagenet'}) #13
#MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'inceptionresnetv2', 'weights': 'imagenet'}) #14
#MODEL_DEFS.append({'arch': 'smp.UnetPlusPlus', 'encoder': 'xception', 'weights': 'imagenet'}) #15
#MODEL_DEFS.append({'arch': 'smp.DeepLabV3Plus', 'encoder': 'xception', 'weights': 'imagenet'}) #16



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_utility.ipynb.

# %% auto 0
__all__ = ['load_cls_labels', 'one_hot_encode', 'reverse_one_hot', 'mask_to_rgb', 'mask_map_classes',
           'mask_code_to_continues_indices', 'mask_continues_indices_to_codes', 'mask_continues_indices_to_rgb',
           'mask_continues_indices_to_idx', 'to_tensor', 'expand_to_32px', 'to_device', 'to_cpu']

# %% ../nbs/01_utility.ipynb 3
import numpy as np
import torch
from PIL import Image
from collections.abc import Mapping

# %% ../nbs/01_utility.ipynb 5
def load_cls_labels(file_path:str):
    "Load class names, codes, and RGB values"
    dtype = {"name": str, "code": int, "color_hex": str}
    cls_dict = pd.read_csv(file_path, dtype=dtype)
    cls_names = cls_dict['name'].tolist()
    cls_code_values = cls_dict['code'].values.tolist()
    
    
    # fill cls_dict with dummy values to allow easy color coding of segmentation masks
    for i in range(int(max(cls_dict['code']))): 
        if cls_dict[cls_dict['code']==i].empty:
            cls_dict.loc[len(cls_dict), 'code'] = i
            cls_dict.loc[len(cls_dict)-1, 'color_hex'] = '000000'
            cls_dict.loc[len(cls_dict)-1, 'name'] = 'dummy'
            cls_dict.loc[len(cls_dict)-1, 'label'] = 'Dummy'
    cls_dict = cls_dict.sort_values('code')  
    cls_color_hex = cls_dict[['color_hex']].values.tolist()
    cls_rgb = np.array([tuple(int(h[0][i:i+2], 16)/255 for i in (0, 2, 4)) for h in cls_color_hex] )
    cls_labels = cls_dict['label'].values.tolist()
    
    #cls_rgb_values = np.nan_to_num(np.array(cls_dict[['r','g','b']].values.tolist()))

    return cls_names, cls_code_values, cls_rgb, cls_labels

# %% ../nbs/01_utility.ipynb 10
def one_hot_encode(label, label_values):
    """
    Convert a segmentation image label array to one-hot format
    by replacing each pixel value with a vector of length num_classes
    # Arguments
        label: The 2D array segmentation image label
        label_values
        
    # Returns
        A 2D array with the same width and height as the input, but
        with a depth size of num_classes
    """
    semantic_map = []
    for colour in label_values:
        equality = np.equal(label, colour)
        class_map = np.all(equality, axis = -1)
        semantic_map.append(class_map)
    semantic_map = np.stack(semantic_map, axis= -1)
    
    return semantic_map

def reverse_one_hot(image):
    """
    Transform a 2D array in one-hot format (depth is num_classes),
    to a 2D array with only 1 channel, where each pixel value is
    the classified class key.
    # Arguments
        image: The one-hot format image 
        
    # Returns
        A 2D array with the same width and hieght as the input, but
        with a depth size of 1, where each pixel value is the classified 
        class key.
    """
    x = np.argmax(image, axis = -1)
    return x



# %% ../nbs/01_utility.ipynb 12
def mask_to_rgb(mask, cls_rgb_values):
    """
    Given a 1-channel array of class keys, color code the segmentation results.
    # Arguments
        mask: single channel array where each value represents the class key.
        cls_rgb_values: np.array of RGB values for keys

    # Returns
        Color coded PIL image for segmentation visualization
    """
    x = cls_rgb_values[mask.astype(int)]
    x = Image.fromarray((255 * x).astype(np.uint8))
    return x

# %% ../nbs/01_utility.ipynb 16
def mask_map_classes(mask, class_mappings):
    if class_mappings is not None:
        for m in class_mappings:
            mask[mask==m[0]]=m[1]
            return mask

# %% ../nbs/01_utility.ipynb 18
def mask_code_to_continues_indices(mask, cls_codes):
    """
    input: mask with n non-continuous class codes
    output: mask with class indices from 0 ... n-1
    """
    m_idx = np.zeros(mask.shape, dtype='uint8')
    for idx, org in enumerate(cls_codes):
        m_idx[mask==org] = idx
    return m_idx

def mask_continues_indices_to_codes(mask, cls_codes):
    return np.array(cls_codes)[mask].astype('uint8')

def mask_continues_indices_to_rgb(mask, cls_codes, cls_rgb_values):
    return mask_to_rgb(mask_continues_indices_to_codes(mask, cls_codes), cls_rgb_values)

# %% ../nbs/01_utility.ipynb 20
def mask_continues_indices_to_idx(mask, select_class_indices):
    return np.array(select_class_indices)[mask].astype('uint8')

# %% ../nbs/01_utility.ipynb 21
def to_tensor(x, **kwargs):
    "convert image (np.array) to float tensor, change order from WHC to CWH"
    if len(x.shape)==2: return x.astype('float32')
    return torch.tensor(x.transpose(2, 0, 1).astype('float32'))

def expand_to_32px(length):
    "return next larger integer divisible by 32"
    return length if length%32 == 0 else length + (32 - length%32)

# %% ../nbs/01_utility.ipynb 23
def to_device(x, device):
    if not device:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if isinstance(x, torch.Tensor): return x.to(device)
    if isinstance(x, Mapping): return {k:v.to(device) for k,v in x.items()}
    return type(x)(to_device(o, device) for o in x)

def to_cpu(x):
    if isinstance(x, Mapping): return {k:to_cpu(v) for k,v in x.items()}
    if isinstance(x, list): return [to_cpu(o) for o in x]
    if isinstance(x, tuple): return tuple(to_cpu(list(x)))
    res = x.detach().cpu()
    return res.float() if res.dtype==torch.float16 else res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_visualization.ipynb.

# %% auto 0
__all__ = ['convert_image', 'show_image_row']

# %% ../nbs/00_visualization.ipynb 2
import matplotlib.pyplot as plt
from torch import Tensor
from torchvision import tv_tensors

# %% ../nbs/00_visualization.ipynb 4
def convert_image(image):
    if isinstance(image, Tensor):
        if image.shape[0] in [1,3]:
            image=image.permute(1,2,0)
    return image

def show_image_row(fontsize=10, figsize=(20,8), **images):
    n_images = len(images)
    plt.figure(figsize=figsize)
    for idx, (name, image) in enumerate(images.items()):
        plt.subplot(1, n_images, idx + 1)
        plt.xticks([])
        plt.yticks([])
        plt.title(name.replace('_',' ').title(), fontsize=fontsize)
        plt.imshow(convert_image(image))
    plt.show()



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/atnclib',
                'doc_host': 'https://atncai.github.io',
                'git_url': 'https://github.com/atncai/atnclib',
                'lib_path': 'atnclib'},
  'syms': { 'atnclib.callbacks': { 'atnclib.callbacks.Callback': ('callbacks.html#callback', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.DeviceCB': ('callbacks.html#devicecb', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.DeviceCB.__init__': ('callbacks.html#devicecb.__init__', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.DeviceCB.before_batch': ( 'callbacks.html#devicecb.before_batch',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.DeviceCB.before_fit': ('callbacks.html#devicecb.before_fit', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB': ('callbacks.html#memorycb', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.after_backprop': ( 'callbacks.html#memorycb.after_backprop',
                                                                                  'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.after_batch': ( 'callbacks.html#memorycb.after_batch',
                                                                               'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.after_epoch': ( 'callbacks.html#memorycb.after_epoch',
                                                                               'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.after_loss': ('callbacks.html#memorycb.after_loss', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.after_preds': ( 'callbacks.html#memorycb.after_preds',
                                                                               'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.before_batch': ( 'callbacks.html#memorycb.before_batch',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.before_epoch': ( 'callbacks.html#memorycb.before_epoch',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.before_fit': ('callbacks.html#memorycb.before_fit', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MemoryCB.before_preds': ( 'callbacks.html#memorycb.before_preds',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB': ('callbacks.html#metricscb', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.__init__': ('callbacks.html#metricscb.__init__', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB._log': ('callbacks.html#metricscb._log', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.after_batch': ( 'callbacks.html#metricscb.after_batch',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.after_epoch': ( 'callbacks.html#metricscb.after_epoch',
                                                                                'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.after_fit': ('callbacks.html#metricscb.after_fit', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.before_epoch': ( 'callbacks.html#metricscb.before_epoch',
                                                                                 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.MetricsCB.before_fit': ( 'callbacks.html#metricscb.before_fit',
                                                                               'atnclib/callbacks.py'),
                                   'atnclib.callbacks.SaveModelCB': ('callbacks.html#savemodelcb', 'atnclib/callbacks.py'),
                                   'atnclib.callbacks.SaveModelCB.__init__': ( 'callbacks.html#savemodelcb.__init__',
                                                                               'atnclib/callbacks.py'),
                                   'atnclib.callbacks.SaveModelCB.after_epoch': ( 'callbacks.html#savemodelcb.after_epoch',
                                                                                  'atnclib/callbacks.py'),
                                   'atnclib.callbacks.get_mem': ('callbacks.html#get_mem', 'atnclib/callbacks.py')},
            'atnclib.core': {'atnclib.core.foo': ('core-copy1.html#foo', 'atnclib/core.py')},
            'atnclib.data': { 'atnclib.data.ATNCDataset': ('data.html#atncdataset', 'atnclib/data.py'),
                              'atnclib.data.ATNCDataset.__getitem__': ('data.html#atncdataset.__getitem__', 'atnclib/data.py'),
                              'atnclib.data.ATNCDataset.__init__': ('data.html#atncdataset.__init__', 'atnclib/data.py'),
                              'atnclib.data.ATNCDataset.__len__': ('data.html#atncdataset.__len__', 'atnclib/data.py'),
                              'atnclib.data.ATNCDataset.get_mask_fns': ('data.html#atncdataset.get_mask_fns', 'atnclib/data.py'),
                              'atnclib.data.DataLoaders': ('data.html#dataloaders', 'atnclib/data.py'),
                              'atnclib.data.DataLoaders.__init__': ('data.html#dataloaders.__init__', 'atnclib/data.py'),
                              'atnclib.data.get_augmentation': ('data.html#get_augmentation', 'atnclib/data.py'),
                              'atnclib.data.get_dataloaders': ('data.html#get_dataloaders', 'atnclib/data.py'),
                              'atnclib.data.get_dataset_fns': ('data.html#get_dataset_fns', 'atnclib/data.py'),
                              'atnclib.data.get_datasets': ('data.html#get_datasets', 'atnclib/data.py'),
                              'atnclib.data.get_preprocessing': ('data.html#get_preprocessing', 'atnclib/data.py'),
                              'atnclib.data.preproc_img': ('data.html#preproc_img', 'atnclib/data.py')},
            'atnclib.learner': { 'atnclib.learner.CancelBatchException': ('learner.html#cancelbatchexception', 'atnclib/learner.py'),
                                 'atnclib.learner.CancelEpochException': ('learner.html#cancelepochexception', 'atnclib/learner.py'),
                                 'atnclib.learner.CancelFitException': ('learner.html#cancelfitexception', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner': ('learner.html#learner', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner.__init__': ('learner.html#learner.__init__', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner.callback': ('learner.html#learner.callback', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner.fit': ('learner.html#learner.fit', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner.one_batch': ('learner.html#learner.one_batch', 'atnclib/learner.py'),
                                 'atnclib.learner.Learner.one_epoch': ('learner.html#learner.one_epoch', 'atnclib/learner.py'),
                                 'atnclib.learner.run_cbs': ('learner.html#run_cbs', 'atnclib/learner.py')},
            'atnclib.losses': { 'atnclib.losses.CombinedLoss': ('losses.html#combinedloss', 'atnclib/losses.py'),
                                'atnclib.losses.CombinedLoss.__call__': ('losses.html#combinedloss.__call__', 'atnclib/losses.py'),
                                'atnclib.losses.CombinedLoss.__init__': ('losses.html#combinedloss.__init__', 'atnclib/losses.py'),
                                'atnclib.losses.CombinedLoss.activation': ('losses.html#combinedloss.activation', 'atnclib/losses.py'),
                                'atnclib.losses.CombinedLoss.decodes': ('losses.html#combinedloss.decodes', 'atnclib/losses.py')},
            'atnclib.metrics': { 'atnclib.metrics.Accuracy': ('metrics.html#accuracy', 'atnclib/metrics.py'),
                                 'atnclib.metrics.Accuracy.__init__': ('metrics.html#accuracy.__init__', 'atnclib/metrics.py'),
                                 'atnclib.metrics.Accuracy.update': ('metrics.html#accuracy.update', 'atnclib/metrics.py'),
                                 'atnclib.metrics.IOU_Score': ('metrics.html#iou_score', 'atnclib/metrics.py'),
                                 'atnclib.metrics.IOU_Score.__init__': ('metrics.html#iou_score.__init__', 'atnclib/metrics.py'),
                                 'atnclib.metrics.IOU_Score.update': ('metrics.html#iou_score.update', 'atnclib/metrics.py'),
                                 'atnclib.metrics.Recall': ('metrics.html#recall', 'atnclib/metrics.py'),
                                 'atnclib.metrics.Recall.__init__': ('metrics.html#recall.__init__', 'atnclib/metrics.py'),
                                 'atnclib.metrics.Recall.update': ('metrics.html#recall.update', 'atnclib/metrics.py'),
                                 'atnclib.metrics.SegmentationMetric': ('metrics.html#segmentationmetric', 'atnclib/metrics.py'),
                                 'atnclib.metrics.SegmentationMetric.__init__': ( 'metrics.html#segmentationmetric.__init__',
                                                                                  'atnclib/metrics.py'),
                                 'atnclib.metrics.SegmentationMetric.compute': ( 'metrics.html#segmentationmetric.compute',
                                                                                 'atnclib/metrics.py'),
                                 'atnclib.metrics.SegmentationMetric.merge_state': ( 'metrics.html#segmentationmetric.merge_state',
                                                                                     'atnclib/metrics.py')},
            'atnclib.model': { 'atnclib.model.calc_confusion_matrix': ('model.html#calc_confusion_matrix', 'atnclib/model.py'),
                               'atnclib.model.create_checkpoint_data': ('model.html#create_checkpoint_data', 'atnclib/model.py'),
                               'atnclib.model.display_confusion_matrix': ('model.html#display_confusion_matrix', 'atnclib/model.py'),
                               'atnclib.model.get_model_name': ('model.html#get_model_name', 'atnclib/model.py'),
                               'atnclib.model.get_preprocessiong_fn': ('model.html#get_preprocessiong_fn', 'atnclib/model.py'),
                               'atnclib.model.get_seg_model': ('model.html#get_seg_model', 'atnclib/model.py'),
                               'atnclib.model.restore_model': ('model.html#restore_model', 'atnclib/model.py'),
                               'atnclib.model.save_model_checkpoint': ('model.html#save_model_checkpoint', 'atnclib/model.py')},
            'atnclib.utility': { 'atnclib.utility.expand_to_32px': ('utility.html#expand_to_32px', 'atnclib/utility.py'),
                                 'atnclib.utility.load_cls_labels': ('utility.html#load_cls_labels', 'atnclib/utility.py'),
                                 'atnclib.utility.mask_code_to_continues_indices': ( 'utility.html#mask_code_to_continues_indices',
                                                                                     'atnclib/utility.py'),
                                 'atnclib.utility.mask_continues_indices_to_codes': ( 'utility.html#mask_continues_indices_to_codes',
                                                                                      'atnclib/utility.py'),
                                 'atnclib.utility.mask_continues_indices_to_idx': ( 'utility.html#mask_continues_indices_to_idx',
                                                                                    'atnclib/utility.py'),
                                 'atnclib.utility.mask_continues_indices_to_rgb': ( 'utility.html#mask_continues_indices_to_rgb',
                                                                                    'atnclib/utility.py'),
                                 'atnclib.utility.mask_map_classes': ('utility.html#mask_map_classes', 'atnclib/utility.py'),
                                 'atnclib.utility.mask_to_rgb': ('utility.html#mask_to_rgb', 'atnclib/utility.py'),
                                 'atnclib.utility.one_hot_encode': ('utility.html#one_hot_encode', 'atnclib/utility.py'),
                                 'atnclib.utility.reverse_one_hot': ('utility.html#reverse_one_hot', 'atnclib/utility.py'),
                                 'atnclib.utility.to_cpu': ('utility.html#to_cpu', 'atnclib/utility.py'),
                                 'atnclib.utility.to_device': ('utility.html#to_device', 'atnclib/utility.py'),
                                 'atnclib.utility.to_tensor': ('utility.html#to_tensor', 'atnclib/utility.py')},
            'atnclib.visualization': { 'atnclib.visualization.convert_image': ( 'visualization.html#convert_image',
                                                                                'atnclib/visualization.py'),
                                       'atnclib.visualization.show_image_row': ( 'visualization.html#show_image_row',
                                                                                 'atnclib/visualization.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/metrics.ipynb.

# %% auto 0
__all__ = ['SegmentationMetric', 'IOU_Score', 'Accuracy', 'Recall']

# %% ../nbs/metrics.ipynb 2
from torcheval.metrics.metric import Metric
import segmentation_models_pytorch as smp
import torch

# %% ../nbs/metrics.ipynb 3
class SegmentationMetric(Metric):
    def __init__(self, num_classes, device=None ) -> None:
        super().__init__(device=device)
        self._add_state('weighted_sum', torch.tensor(0.0, device=self.device, dtype=torch.float64))
        self._add_state('weights', torch.tensor(0.0, device=self.device, dtype=torch.float64))
        self.num_classes = num_classes
        
    @torch.inference_mode()
    def compute(self):
        return self.weighted_sum/self.weights
    
    @torch.inference_mode()
    def merge_state(self, metrics):
        for metric in metrics:
            self.weighted_sum += metric.weighted_sum.to(self.device)
            self.weights += metric.weights.to(self.device)
        return self


# %% ../nbs/metrics.ipynb 4
class IOU_Score(SegmentationMetric):
    def __init__(self, num_classes, device=None, reduction='micro') -> None:
        super().__init__(num_classes=num_classes, device=device)
        self.reduction = reduction
    
    @torch.inference_mode()
    def update(self, preds, target):
        if len(preds.shape)==4:
            preds = preds.softmax(dim=1).argmax(dim=1)
        tp, fp, fn, tn = smp.metrics.get_stats(preds, target, mode='multiclass', num_classes=self.num_classes)
        self.weighted_sum += smp.metrics.iou_score(tp, fp, fn, tn, reduction=self.reduction)
        self.weights += 1# tp.shape[0]
        return self
    

    
class Accuracy(SegmentationMetric):
    def __init__(self, num_classes, device=None, reduction='macro') -> None:
        super().__init__(num_classes=num_classes, device=device)
        self.reduction = reduction

    @torch.inference_mode()
    def update(self, preds, target):
        if len(preds.shape)==4:
            preds = preds.softmax(dim=1).argmax(dim=1)
        tp, fp, fn, tn = smp.metrics.get_stats(preds, target, mode='multiclass', num_classes=self.num_classes)
        self.weighted_sum += smp.metrics.accuracy(tp, fp, fn, tn, reduction=self.reduction)
        self.weights += 1# tp.shape[0]
        return self


class Recall(SegmentationMetric):
    def __init__(self, num_classes, device=None, reduction='micro-imagewise') -> None:
        super().__init__(num_classes=num_classes, device=device)
        self.reduction = reduction

    @torch.inference_mode()
    def update(self, preds, target):
        if len(preds.shape)==4:
            preds = preds.softmax(dim=1).argmax(dim=1)
        tp, fp, fn, tn = smp.metrics.get_stats(preds, target, mode='multiclass', num_classes=self.num_classes)
        self.weighted_sum += smp.metrics.recall(tp, fp, fn, tn, reduction=self.reduction)
        self.weights += 1# tp.shape[0]
        return self




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Callbacks.ipynb.

# %% auto 0
__all__ = ['Callback', 'MetricsCB', 'DeviceCB', 'get_mem', 'MemoryCB', 'SaveModelCB']

# %% ../nbs/Callbacks.ipynb 2
import torch
from .utility import to_cpu, to_device
from copy import copy
from torcheval.metrics import Mean
from torch.utils.tensorboard import SummaryWriter

# %% ../nbs/Callbacks.ipynb 3
class Callback(): order = 0

# %% ../nbs/Callbacks.ipynb 4
class MetricsCB(Callback):
    def __init__(self, device=None, use_tensorboard=False, *ms, **metrics):
        for o in ms: metrics[type(o).__name__] = o
        self.metrics = metrics
        self.all_metrics = copy(metrics)
        self.all_metrics['loss'] = self.loss = Mean(device=device)
        self.use_tensorboard = use_tensorboard

    def _log(self, d): print(d)
    def before_fit(self, learn): 
        learn.metrics = self
        if self.use_tensorboard:
            self.writer = SummaryWriter()
            
    def before_epoch(self, learn): [o.reset() for o in self.all_metrics.values()]

    def after_epoch(self, learn):
        log = {k:f'{v.compute():.3f}' for k,v in self.all_metrics.items()}
        
        if learn.model.training:
            epoch_type = 'train'  
        else: 
            epoch_type = 'valid'
            learn.current_metric_values = log
            offset = 0.5
        log['epoch'] = learn.epoch
        log['train'] = epoch_type
        self._log(log)
        if self.use_tensorboard:
            for k, v in self.all_metrics.items():
                self.writer.add_scalars('data/'+ k, {epoch_type:v.compute()}, learn.epoch)
            
    def after_batch(self, learn):
        #x,y,*_ = to_cpu(learn.batch)
        x,y,*_ = learn.batch
        #for m in self.metrics.values(): m.update(to_cpu(learn.preds), y)
        #loss = to_cpu(learn.loss)
        for m in self.metrics.values(): m.update(learn.preds, y)
        
        loss = learn.loss
        self.loss.update(loss, weight=len(x))
        
    def after_fit(self, learn):
        if self.use_tensorboard:
            self.writer.flush()
            self.writer.close()

# %% ../nbs/Callbacks.ipynb 5
class DeviceCB(Callback):
    def __init__(self, device=None): 
        if not device:
            device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.device = device
    def before_fit(self, learn):
        if hasattr(learn.model, 'to'): 
            learn.model.to(self.device)
    def before_batch(self, learn): 
        learn.batch = to_device(learn.batch, device=self.device)    

# %% ../nbs/Callbacks.ipynb 6
def get_mem():
    return torch.cuda.memory_allocated()/1024/1024

class MemoryCB(Callback):
    def before_fit(self, learn):
        print(f'memory befor fit {get_mem()}')
    def before_batch(self, learn): 
        print(f'memory before batch {get_mem()}')
    def after_batch(self, learn): 
        print(f'memory after batch {get_mem()}')
    def before_epoch(self, learn): 
        print(f'memory before epoch {get_mem()}')
    def after_epoch(self, learn): 
        print(f'memory after epoch {get_mem()}')
    def after_loss(self, learn): 
        print(f'memory after loss {get_mem()}')
    def after_preds(self, learn): 
        print(f'memory after preds {get_mem()}')
    def before_preds(self, learn): 
        print(f'memory before preds{get_mem()}')
    def after_backprop(self, learn): 
        print(f'memory after backprop {get_mem()}')

# %% ../nbs/Callbacks.ipynb 7
from .model import save_model_checkpoint, create_checkpoint_data

class SaveModelCB(Callback):
    order = 9
    
    def __init__(self, metric, metric_limit, model_path, model_name, selected_classes, model_def):
        self.metric = metric
        self.metric_limit = int(1000 * metric_limit)
        self.best_metric_value = 0.0
        self.model_path = model_path
        self.model_name = model_name
        self.selected_classes = selected_classes
        self.model_def = model_def
        self.best_model_name = model_name + '_best_model.pth'
    
    def after_epoch(self, learn):
        if learn.model.training: return #  only save after validation epoch
        values = learn.current_metric_values
        if not values: return
        
        try:    
            current_metric_value = float(values[self.metric])
        except:
            print('metric as save criterion not in metrics')
            return 
        epoch = values['epoch']
        loss = values['loss']

        scaled_metric = int(1000 * current_metric_value)
        if self.best_metric_value < current_metric_value:
            self.best_metric_value = current_metric_value
            checkpoint_data = create_checkpoint_data(model_def = self.model_def,
                                                     select_classes=self.selected_classes,
                                                     epoch = epoch,
                                                     loss = loss,
                                                     metric = current_metric_value)
            
            save_model_checkpoint(learn.model, learn.opt, self.model_path/self.best_model_name, **checkpoint_data)
            print(f'Best model saved with {self.metric} = {current_metric_value:.3}')
            if scaled_metric > self.metric_limit:
                numbered_model_name = self.model_name + '_' + self.metric + f'{scaled_metric:03}' + '.pth'
                save_model_checkpoint(learn.model, learn.opt, self.model_path / numbered_model_name, **checkpoint_data)
                print(f'Numbered Model saved with {self.metric} = {current_metric_value:.3}')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_learner.ipynb.

# %% auto 0
__all__ = ['run_cbs', 'CancelFitException', 'CancelBatchException', 'CancelEpochException', 'Learner']

# %% ../nbs/05_learner.ipynb 3
import torch
from contextlib import contextmanager
from operator import attrgetter
from torch import optim
import torch.nn.functional as F
from functools import partial

from .callbacks import Callback
from .model import get_seg_model, get_model_name

# %% ../nbs/05_learner.ipynb 4
def run_cbs(cbs, method_nm, learn=None):
    for cb in sorted(cbs, key=attrgetter('order')):
        method = getattr(cb, method_nm, None)
        
        if method is not None: 
            method(learn)

# %% ../nbs/05_learner.ipynb 7
class CancelFitException(Exception): pass
class CancelBatchException(Exception): pass
class CancelEpochException(Exception): pass

# %% ../nbs/05_learner.ipynb 8
class Learner():
    def __init__(self, model, dls, cbs, loss_func=F.cross_entropy, lr=0.1, opt_func=optim.SGD):
        self.model = model
        self.dls = dls
        self.loss_func = loss_func
        self.opt_func = opt_func
        self.cbs = cbs
        self.n_epochs = 1
        self.epochs = None
        self.opt = None
        self.dl = None
        self.lr = lr

    def one_batch(self):
        self.callback('before_preds')
        self.preds = self.model(self.batch[0])
        self.callback('after_preds')
        self.loss = self.loss_func(self.preds, self.batch[1])
        self.callback('after_loss')
        if self.model.training:
            self.loss.backward()
            self.opt.step()
            self.opt.zero_grad()
        self.callback('after_backprop')

    def one_epoch(self, train):
        self.model.train(train)
        self.dl = self.dls.train if train else self.dls.valid
        try:
            self.callback('before_epoch')
            for self.iter,self.batch in enumerate(self.dl):
                try:
                    self.callback('before_batch')
                    self.one_batch()
                    self.callback('after_batch')
                except CancelBatchException: pass
            self.callback('after_epoch')
        except CancelEpochException: pass
    
    def fit(self, n_epochs):
        self.n_epochs = n_epochs
        self.epochs = range(n_epochs)
        self.opt = self.opt_func(self.model.parameters(), self.lr)
        try:
            self.callback('before_fit')
            for self.epoch in self.epochs:
                self.callback('before_full_epoch')
                self.one_epoch(True)
                self.one_epoch(False)
                self.callback('after_full_epoch')
            self.callback('after_fit')
        except CancelFitException: pass

    def callback(self, method_nm): run_cbs(self.cbs, method_nm, self)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_data.ipynb.

# %% auto 0
__all__ = ['get_dataset_fns', 'get_preprocessing', 'get_augmentation', 'ATNCDataset', 'DataLoaders', 'get_datasets',
           'get_dataloaders', 'preproc_img']

# %% ../nbs/02_data.ipynb 3
from pathlib import Path
import random
import albumentations as album
from torch.utils.data import Dataset, DataLoader
import numpy as np
from .utility import to_tensor, mask_map_classes, mask_code_to_continues_indices, load_cls_labels, mask_continues_indices_to_codes, mask_to_rgb
from .visualization import *
from .model import *
from PIL import Image

# %% ../nbs/02_data.ipynb 5
def get_dataset_fns(path, data_dirs, img_dirs, seed=42, ext='png', dist=[0.8,0.1,0.1], img_count=-1):
    """
    Returns 3 randomized lists of filenames for train, val, and test datasets
    """
    all_fns = []
    for data_dir in data_dirs:
        for img_dir in img_dirs:
            img_path = path/data_dir/img_dir
            all_fns += sorted(list(img_path.glob('*.'+ext)))

    random.Random(seed).shuffle(all_fns)
    if img_count == -1:
        img_count = len(all_fns)
    train_count = round(img_count * dist[0])
    val_count = round(img_count * dist[1])
    test_count = img_count - train_count - val_count
        
    train_fns = all_fns[0:train_count]
    val_fns = all_fns[train_count:val_count+train_count]
    test_fns = all_fns[train_count + val_count:train_count + val_count + test_count]
    
    return train_fns, val_fns, test_fns
                


# %% ../nbs/02_data.ipynb 8
def get_preprocessing(preprocessing_fn=None):
    """Construct preprocessing transform    
    Args:
        preprocessing_fn (callable): data normalization function 
            (can be specific for each pretrained neural network)
    Return:
        transform: albumentations.Compose
    """ 
    _transform = []
    if preprocessing_fn:
        _transform.append(album.Lambda(image=preprocessing_fn))
    _transform.append(album.Lambda(image=to_tensor, mask=to_tensor))
    
    return album.Compose(_transform)

# %% ../nbs/02_data.ipynb 10
def get_augmentation(transforms):
    """Construct augmentation transform
    Args:
        list of augmentation transforms
    Return:
        transform: albumentations.Compose
    """
    return album.Compose(transforms)

# %% ../nbs/02_data.ipynb 13
class ATNCDataset(Dataset):
    """Dataset. Read images, apply augmentation and preprocessing transformations.
    
    Args:
        image_fns ([Paths]): paths to images folder
        mask_path (str): path to segmentation masks folder
        cls_codes (list): class codes of select classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transfromation pipeline 
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing 
            (e.g. noralization, shape manipulation, etc.)
    
    """
    
    def __init__(
            self,
            image_fns,
            mask_path,
            cls_codes=None,
            class_mappings=None,
            augmentation=None,
            preprocessing=None,
            is_org=False
            ):
        self.image_fns = image_fns
        self.mask_fns = self.get_mask_fns(image_fns, mask_path)
        self.cls_codes = cls_codes
        self.augmentation = augmentation
        self.preprocessing = preprocessing
        self.is_org=is_org
        self.class_mappings = class_mappings
    
    def __getitem__(self, i):
        image = np.array(Image.open(str(self.image_fns[i])))
        if self.is_org:
            # return empty mask if original (not annotated) image
            mask = np.zeros(image.shape)
        else:
            mask = np.array(Image.open(str(self.mask_fns[i])))#[:,:,0]

        if self.class_mappings is not None:
            mask = mask_map_classes(mask, self.class_mappings)
        if self.cls_codes is not None:
           mask = mask_code_to_continues_indices(mask, self.cls_codes)
        
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample['image'], sample['mask']
            
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample['image'], sample['mask']
        
        return image, mask.astype('int')
    
    def __len__(self):
        return len(self.image_fns)

    #def get_mask_fns(self, image_fns, mask_path):
    #    return [o.parent.parent/mask_path/('seg_'+o.name.split('_',1)[-1]) for o in image_fns]
    
    def get_mask_fns(self, image_fns, mask_path):
        mask_fns = []
        for o in image_fns:
            if 'img' in o.name:
                source_name = 'img'
            else:
                source_name = 'noisyImg'
            
            mask_fn = o.parent.parent/mask_path/(o.name.replace(source_name,'seg'))
            if mask_fn.exists() == False:
                mask_name = mask_fn.name.split('.png')[0] + '.tif'
                mask_fn = mask_fn.parent/mask_name
            mask_fns.append(mask_fn)    
        return mask_fns #[o.parent.parent/mask_path/(o.name.replace('img','seg')) for o in image_fns]

# %% ../nbs/02_data.ipynb 15
class DataLoaders():
    def __init__(self, train_dl, valid_dl):
        self.train = train_dl
        self.valid= valid_dl

def get_datasets(train_fns, val_fns, masks_folder, cls_codes, class_mappings, train_augmentation, valid_augmentation, model_def):
    preprocessing = get_preprocessing(get_preprocessiong_fn(model_def))
    train_ds = ATNCDataset(image_fns=train_fns,
                           mask_path=masks_folder,
                           cls_codes=cls_codes,
                           class_mappings=class_mappings,
                           augmentation=train_augmentation,
                           preprocessing=preprocessing,
                           is_org=False)
    valid_ds = ATNCDataset(image_fns=val_fns,
                           mask_path=masks_folder,
                           cls_codes=cls_codes,
                           class_mappings=class_mappings,
                           augmentation=valid_augmentation,
                           preprocessing=preprocessing,
                           is_org=False)
    return train_ds, valid_ds

def get_dataloaders(train_ds, valid_ds, train_bs=4,valid_bs = 4, num_workers = 4):
    train_dl = DataLoader(train_ds, batch_size=train_bs, shuffle=True, num_workers=num_workers)
    valid_dl = DataLoader(valid_ds, batch_size=valid_bs, shuffle=False, num_workers=num_workers)
    return DataLoaders(train_dl, valid_dl)

# %% ../nbs/02_data.ipynb 17
def preproc_img(img, model_def, aug_transform=[]):
    preproc = get_preprocessiong_fn(model_def)
    transform = [
        album.PadIfNeeded(min_height = None, min_width = None, pad_height_divisor=32, pad_width_divisor=32, always_apply=True),
            ] + aug_transform
    aug = get_augmentation(transform)
    tmp_img = aug(image = img)['image']
    tmp_img = preproc(tmp_img)
    
    return to_tensor(tmp_img).unsqueeze(dim=0)
